from azure.ai.ml.constants._common import AssetTypes
from azure.ai.ml.entities._inputs_outputs import Input, Output
from typing import Optional, Dict
from azure.ai.ml._restclient.v2024_10_01_preview.models import (
    UriFileJobInput,
    MLFlowModelJobInput,
)
from azure.ai.ml.constants._job.finetuning import FineTuningTaskTypes
from azure.ai.ml.entities._job.finetuning.custom_model_finetuning_job import (
    CustomModelFineTuningJob,
)
from azure.ai.ml.entities._job.finetuning.azure_openai_finetuning_job import (
    AzureOpenAIFineTuningJob,
)
from azure.ai.ml.entities._job.finetuning.azure_openai_hyperparameters import (
    AzureOpenAIHyperparameters,
)
import pytest


@pytest.mark.finetuning_test
@pytest.mark.unittest
class TestCustomModelFineTuningJob:
    @pytest.mark.parametrize(
        "task",
        [
            FineTuningTaskTypes.CHAT_COMPLETION,
            FineTuningTaskTypes.TEXT_COMPLETION,
            FineTuningTaskTypes.TEXT_CLASSIFICATION,
            FineTuningTaskTypes.QUESTION_ANSWERING,
            FineTuningTaskTypes.TEXT_SUMMARIZATION,
            FineTuningTaskTypes.TOKEN_CLASSIFICATION,
            FineTuningTaskTypes.TEXT_TRANSLATION,
            FineTuningTaskTypes.IMAGE_CLASSIFICATION,
            FineTuningTaskTypes.IMAGE_INSTANCE_SEGMENTATION,
            FineTuningTaskTypes.IMAGE_OBJECT_DETECTION,
            FineTuningTaskTypes.VIDEO_MULTI_OBJECT_TRACKING,
        ],
    )
    def test_custom_model_finetuning_job_conversion(self, task: str):
        # Create Custom Model FineTuning Job
        custom_model_finetuning_job = self._get_custom_model_finetuning_job(
            task=task,
            display_name="llama-display-name",
            training_data=Input(type=AssetTypes.URI_FILE, path="https://foo/bar/train.csv"),
            validation_data=Input(type=AssetTypes.URI_FILE, path="https://foo/bar/test.csv"),
            hyperparameters={"foo": "bar"},
            model=Input(
                type=AssetTypes.MLFLOW_MODEL,
                path="azureml://registries/azureml-meta/models/Llama-2-7b/versions/9",
            ),
            name="llama-finetuning",
            experiment_name="foo_exp",
            tags={"foo_tag": "bar"},
            properties={"my_property": True},
            outputs={"registered_model": Output(type="mlflow_model", name="llama-finetune-registered")},
        )
        rest_obj = custom_model_finetuning_job._to_rest_object()
        assert isinstance(
            rest_obj.properties.fine_tuning_details.model, MLFlowModelJobInput
        ), "Model is not MLFlowModelJobInput"
        assert isinstance(
            rest_obj.properties.fine_tuning_details.training_data, UriFileJobInput
        ), "Training data is not UriFileJobInput"
        assert isinstance(
            rest_obj.properties.fine_tuning_details.validation_data, UriFileJobInput
        ), "Validation data is not UriFileJobInput"

        original_obj = CustomModelFineTuningJob._from_rest_object(rest_obj)
        assert custom_model_finetuning_job == original_obj, "Conversion to/from rest object failed"
        assert original_obj.task.lower() == task.lower(), "Task not set correctly"
        assert original_obj.display_name == "llama-display-name", "Display name not set correctly"
        assert original_obj.name == "llama-finetuning", "Name not set correctly"
        assert original_obj.experiment_name == "foo_exp", "Experiment name not set correctly"
        assert original_obj.tags == {"foo_tag": "bar"}, "Tags not set correctly"
        assert original_obj.properties == {"my_property": True}, "Properties not set correctly"
        # check if the original job inputs were restored
        assert isinstance(original_obj.training_data, Input), "Training data is not Input"
        assert original_obj.training_data.type == AssetTypes.URI_FILE, "Training data type not set correctly"
        assert original_obj.training_data.path == "https://foo/bar/train.csv", "Training data path not set correctly"
        assert isinstance(original_obj.validation_data, Input), "Test data is not Input"
        assert original_obj.validation_data.type == AssetTypes.URI_FILE, "Test data type not set correctly"
        assert original_obj.validation_data.path == "https://foo/bar/test.csv", "Test data path not set correctly"
        assert original_obj.hyperparameters == {"foo": "bar"}, "Hyperparameters not set correctly"
        assert isinstance(original_obj.model, Input), "Model is not Input"
        assert original_obj.model.type == AssetTypes.MLFLOW_MODEL, "Model type not set correctly"
        assert (
            original_obj.model.path == "azureml://registries/azureml-meta/models/Llama-2-7b/versions/9"
        ), "Model path not set correctly"

        assert isinstance(original_obj.outputs["registered_model"], Output), "Output is not Output"
        mlflow_model_output = original_obj.outputs["registered_model"]
        assert mlflow_model_output.type == "mlflow_model", "Output type not set correctly"
        assert mlflow_model_output.name == "llama-finetune-registered", "Output name not set correctly"

    @pytest.mark.parametrize(
        "task",
        [
            FineTuningTaskTypes.CHAT_COMPLETION,
            FineTuningTaskTypes.TEXT_COMPLETION,
            FineTuningTaskTypes.TEXT_CLASSIFICATION,
            FineTuningTaskTypes.QUESTION_ANSWERING,
            FineTuningTaskTypes.TEXT_SUMMARIZATION,
            FineTuningTaskTypes.TOKEN_CLASSIFICATION,
            FineTuningTaskTypes.TEXT_TRANSLATION,
            FineTuningTaskTypes.IMAGE_CLASSIFICATION,
            FineTuningTaskTypes.IMAGE_INSTANCE_SEGMENTATION,
            FineTuningTaskTypes.IMAGE_OBJECT_DETECTION,
            FineTuningTaskTypes.VIDEO_MULTI_OBJECT_TRACKING,
        ],
    )
    def test_custom_model_finetuning_job_read_from_wire(self, task: str):
        custom_model_finetuning_job = self._get_custom_model_finetuning_job(
            task=task,
            display_name="llama-display-name",
            training_data=Input(type=AssetTypes.URI_FILE, path="./samsum_dataset/small_train.jsonl"),
            validation_data=Input(type=AssetTypes.URI_FILE, path="./samsum_dataset/small_validation.jsonl"),
            hyperparameters={"foo": "bar"},
            model=Input(
                type=AssetTypes.MLFLOW_MODEL,
                path="azureml://registries/azureml-meta/models/Llama-2-7b/versions/9",
            ),
            name="llama-finetuning",
            experiment_name="foo_exp",
            tags={"foo_tag": "bar"},
            properties={"my_property": True},
            outputs={"registered_model": Output(type="mlflow_model", name="llama-finetune-registered")},
        )
        dict_obj = custom_model_finetuning_job._to_dict()
        assert dict_obj["task"].lower() == task.lower(), "Task not set correctly"
        assert dict_obj["display_name"] == "llama-display-name", "Display name not set correctly"
        assert dict_obj["name"] == "llama-finetuning", "Name not set correctly"
        assert dict_obj["experiment_name"] == "foo_exp", "Experiment name not set correctly"
        assert dict_obj["tags"] == {"foo_tag": "bar"}, "Tags not set correctly"
        assert dict_obj["properties"]["my_property"] == "True", "Properties not set correctly"
        # check if the original job inputs were restored
        assert dict_obj["training_data"]["type"] == AssetTypes.URI_FILE, "Training data type not set correctly"
        assert (
            dict_obj["training_data"]["path"] == "azureml:./samsum_dataset/small_train.jsonl"
        ), "Training data path not set correctly"
        assert dict_obj["validation_data"]["type"] == AssetTypes.URI_FILE, "validation data type not set correctly"
        assert (
            dict_obj["validation_data"]["path"] == "azureml:./samsum_dataset/small_validation.jsonl"
        ), "Validation data path not set correctly"
        assert dict_obj["hyperparameters"] == {"foo": "bar"}, "Hyperparameters not set correctly"
        assert dict_obj["model"]["type"] == AssetTypes.MLFLOW_MODEL, "Model type not set correctly"
        assert (
            dict_obj["model"]["path"] == "azureml://registries/azureml-meta/models/Llama-2-7b/versions/9"
        ), "Model path not set correctly"

    @pytest.mark.parametrize(
        "task",
        [
            FineTuningTaskTypes.CHAT_COMPLETION,
            FineTuningTaskTypes.TEXT_COMPLETION,
            FineTuningTaskTypes.TEXT_CLASSIFICATION,
            FineTuningTaskTypes.QUESTION_ANSWERING,
            FineTuningTaskTypes.TEXT_SUMMARIZATION,
            FineTuningTaskTypes.TOKEN_CLASSIFICATION,
            FineTuningTaskTypes.TEXT_TRANSLATION,
            FineTuningTaskTypes.IMAGE_CLASSIFICATION,
            FineTuningTaskTypes.IMAGE_INSTANCE_SEGMENTATION,
            FineTuningTaskTypes.IMAGE_OBJECT_DETECTION,
            FineTuningTaskTypes.VIDEO_MULTI_OBJECT_TRACKING,
        ],
    )
    def test_azure_openai_finetuning_job_conversion(self, task: str):
        # Create Custom Model FineTuning Job
        custom_model_finetuning_job = self._get_azure_openai_finetuning_job(
            task=task,
            training_data=Input(type=AssetTypes.URI_FILE, path="https://foo/bar/train.csv"),
            validation_data=Input(type=AssetTypes.URI_FILE, path="https://foo/bar/test.csv"),
            hyperparameters=AzureOpenAIHyperparameters(batch_size=2, n_epochs=3, learning_rate_multiplier=0.5),
            model=Input(
                type=AssetTypes.MLFLOW_MODEL,
                path="azureml://registries/azure-openai-v2/models/gpt-4/versions/1",
            ),
            name="gpt4-finetuning",
            experiment_name="foo_exp",
            tags={"foo_tag": "bar"},
            properties={"my_property": True},
        )
        rest_obj = custom_model_finetuning_job._to_rest_object()
        assert isinstance(
            rest_obj.properties.fine_tuning_details.model, MLFlowModelJobInput
        ), "Model is not MLFlowModelJobInput"
        assert isinstance(
            rest_obj.properties.fine_tuning_details.training_data, UriFileJobInput
        ), "Training data is not UriFileJobInput"
        assert isinstance(
            rest_obj.properties.fine_tuning_details.validation_data, UriFileJobInput
        ), "Validation data is not UriFileJobInput"

        original_obj = AzureOpenAIFineTuningJob._from_rest_object(rest_obj)
        assert custom_model_finetuning_job == original_obj, "Conversion to/from rest object failed"
        assert original_obj.task.lower() == task.lower(), "Task not set correctly"
        assert original_obj.name == "gpt4-finetuning", "Name not set correctly"
        assert original_obj.experiment_name == "foo_exp", "Experiment name not set correctly"
        assert original_obj.tags == {"foo_tag": "bar"}, "Tags not set correctly"
        assert original_obj.properties == {"my_property": True}, "Properties not set correctly"
        # check if the original job inputs were restored
        assert isinstance(original_obj.training_data, Input), "Training data is not Input"
        assert original_obj.training_data.type == AssetTypes.URI_FILE, "Training data type not set correctly"
        assert original_obj.training_data.path == "https://foo/bar/train.csv", "Training data path not set correctly"
        assert isinstance(original_obj.validation_data, Input), "Test data is not Input"
        assert original_obj.validation_data.type == AssetTypes.URI_FILE, "Test data type not set correctly"
        assert original_obj.validation_data.path == "https://foo/bar/test.csv", "Test data path not set correctly"
        assert original_obj.hyperparameters.batch_size == 2, "Batch size not set correctly"
        assert original_obj.hyperparameters.n_epochs == 3, "n_epochs not set correctly"
        assert (
            original_obj.hyperparameters.learning_rate_multiplier == 0.5
        ), "learning_rate_multiplier not set correctly"
        assert isinstance(original_obj.model, Input), "Model is not Input"
        assert original_obj.model.type == AssetTypes.MLFLOW_MODEL, "Model type not set correctly"
        assert (
            original_obj.model.path == "azureml://registries/azure-openai-v2/models/gpt-4/versions/1"
        ), "Model path not set correctly"

    def _get_custom_model_finetuning_job(
        self,
        *,
        model: Input,
        task: str,
        training_data: Input,
        validation_data: Optional[Input] = None,
        hyperparameters: Optional[Dict[str, str]] = None,
        **kwargs,
    ) -> CustomModelFineTuningJob:
        custom_model_finetuning_job = CustomModelFineTuningJob(
            task=task,
            model=model,
            training_data=training_data,
            validation_data=validation_data,
            hyperparameters=hyperparameters,
            **kwargs,
        )
        return custom_model_finetuning_job

    def _get_azure_openai_finetuning_job(
        self,
        *,
        model: Input,
        task: str,
        training_data: Input,
        validation_data: Optional[Input] = None,
        hyperparameters: Optional[AzureOpenAIHyperparameters] = None,
        **kwargs,
    ) -> AzureOpenAIFineTuningJob:
        custom_model_finetuning_job = AzureOpenAIFineTuningJob(
            task=task,
            model=model,
            training_data=training_data,
            validation_data=validation_data,
            hyperparameters=hyperparameters,
            **kwargs,
        )
        return custom_model_finetuning_job



# coding=utf-8
# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# Code generated by Microsoft (R) AutoRest Code Generator.
# Changes may cause incorrect behavior and will be lost if the code is regenerated.
# --------------------------------------------------------------------------

from copy import deepcopy
from typing import TYPE_CHECKING

from azure.mgmt.core import ARMPipelineClient
from msrest import Deserializer, Serializer

from . import models
from ._configuration import AzureMachineLearningWorkspacesConfiguration
from .operations import BatchDeploymentsOperations, BatchEndpointsOperations, CodeContainersOperations, CodeVersionsOperations, ComponentContainersOperations, ComponentVersionsOperations, ComputeOperations, DataContainersOperations, DataVersionsOperations, DatastoresOperations, EnvironmentContainersOperations, EnvironmentVersionsOperations, JobsOperations, ModelContainersOperations, ModelVersionsOperations, OnlineDeploymentsOperations, OnlineEndpointsOperations, Operations, PrivateEndpointConnectionsOperations, PrivateLinkResourcesOperations, QuotasOperations, UsagesOperations, VirtualMachineSizesOperations, WorkspaceConnectionsOperations, WorkspaceFeaturesOperations, WorkspacesOperations

if TYPE_CHECKING:
    # pylint: disable=unused-import,ungrouped-imports
    from typing import Any, Optional

    from azure.core.credentials import TokenCredential
    from azure.core.rest import HttpRequest, HttpResponse

class AzureMachineLearningWorkspaces(object):
    """These APIs allow end users to operate on Azure Machine Learning Workspace resources.

    :ivar batch_endpoints: BatchEndpointsOperations operations
    :vartype batch_endpoints:
     azure.mgmt.machinelearningservices.operations.BatchEndpointsOperations
    :ivar batch_deployments: BatchDeploymentsOperations operations
    :vartype batch_deployments:
     azure.mgmt.machinelearningservices.operations.BatchDeploymentsOperations
    :ivar code_containers: CodeContainersOperations operations
    :vartype code_containers:
     azure.mgmt.machinelearningservices.operations.CodeContainersOperations
    :ivar code_versions: CodeVersionsOperations operations
    :vartype code_versions: azure.mgmt.machinelearningservices.operations.CodeVersionsOperations
    :ivar component_containers: ComponentContainersOperations operations
    :vartype component_containers:
     azure.mgmt.machinelearningservices.operations.ComponentContainersOperations
    :ivar component_versions: ComponentVersionsOperations operations
    :vartype component_versions:
     azure.mgmt.machinelearningservices.operations.ComponentVersionsOperations
    :ivar data_containers: DataContainersOperations operations
    :vartype data_containers:
     azure.mgmt.machinelearningservices.operations.DataContainersOperations
    :ivar data_versions: DataVersionsOperations operations
    :vartype data_versions: azure.mgmt.machinelearningservices.operations.DataVersionsOperations
    :ivar datastores: DatastoresOperations operations
    :vartype datastores: azure.mgmt.machinelearningservices.operations.DatastoresOperations
    :ivar environment_containers: EnvironmentContainersOperations operations
    :vartype environment_containers:
     azure.mgmt.machinelearningservices.operations.EnvironmentContainersOperations
    :ivar environment_versions: EnvironmentVersionsOperations operations
    :vartype environment_versions:
     azure.mgmt.machinelearningservices.operations.EnvironmentVersionsOperations
    :ivar jobs: JobsOperations operations
    :vartype jobs: azure.mgmt.machinelearningservices.operations.JobsOperations
    :ivar model_containers: ModelContainersOperations operations
    :vartype model_containers:
     azure.mgmt.machinelearningservices.operations.ModelContainersOperations
    :ivar model_versions: ModelVersionsOperations operations
    :vartype model_versions: azure.mgmt.machinelearningservices.operations.ModelVersionsOperations
    :ivar online_endpoints: OnlineEndpointsOperations operations
    :vartype online_endpoints:
     azure.mgmt.machinelearningservices.operations.OnlineEndpointsOperations
    :ivar online_deployments: OnlineDeploymentsOperations operations
    :vartype online_deployments:
     azure.mgmt.machinelearningservices.operations.OnlineDeploymentsOperations
    :ivar operations: Operations operations
    :vartype operations: azure.mgmt.machinelearningservices.operations.Operations
    :ivar workspaces: WorkspacesOperations operations
    :vartype workspaces: azure.mgmt.machinelearningservices.operations.WorkspacesOperations
    :ivar usages: UsagesOperations operations
    :vartype usages: azure.mgmt.machinelearningservices.operations.UsagesOperations
    :ivar virtual_machine_sizes: VirtualMachineSizesOperations operations
    :vartype virtual_machine_sizes:
     azure.mgmt.machinelearningservices.operations.VirtualMachineSizesOperations
    :ivar quotas: QuotasOperations operations
    :vartype quotas: azure.mgmt.machinelearningservices.operations.QuotasOperations
    :ivar compute: ComputeOperations operations
    :vartype compute: azure.mgmt.machinelearningservices.operations.ComputeOperations
    :ivar private_endpoint_connections: PrivateEndpointConnectionsOperations operations
    :vartype private_endpoint_connections:
     azure.mgmt.machinelearningservices.operations.PrivateEndpointConnectionsOperations
    :ivar private_link_resources: PrivateLinkResourcesOperations operations
    :vartype private_link_resources:
     azure.mgmt.machinelearningservices.operations.PrivateLinkResourcesOperations
    :ivar workspace_connections: WorkspaceConnectionsOperations operations
    :vartype workspace_connections:
     azure.mgmt.machinelearningservices.operations.WorkspaceConnectionsOperations
    :ivar workspace_features: WorkspaceFeaturesOperations operations
    :vartype workspace_features:
     azure.mgmt.machinelearningservices.operations.WorkspaceFeaturesOperations
    :param credential: Credential needed for the client to connect to Azure.
    :type credential: ~azure.core.credentials.TokenCredential
    :param subscription_id: The ID of the target subscription.
    :type subscription_id: str
    :param base_url: Service URL. Default value is 'https://management.azure.com'.
    :type base_url: str
    :keyword int polling_interval: Default waiting time between two polls for LRO operations if no
     Retry-After header is present.
    """

    def __init__(
        self,
        credential,  # type: "TokenCredential"
        subscription_id,  # type: str
        base_url="https://management.azure.com",  # type: str
        **kwargs  # type: Any
    ):
        # type: (...) -> None
        self._config = AzureMachineLearningWorkspacesConfiguration(credential=credential, subscription_id=subscription_id, **kwargs)
        self._client = ARMPipelineClient(base_url=base_url, config=self._config, **kwargs)

        client_models = {k: v for k, v in models.__dict__.items() if isinstance(v, type)}
        self._serialize = Serializer(client_models)
        self._deserialize = Deserializer(client_models)
        self._serialize.client_side_validation = False
        self.batch_endpoints = BatchEndpointsOperations(self._client, self._config, self._serialize, self._deserialize)
        self.batch_deployments = BatchDeploymentsOperations(self._client, self._config, self._serialize, self._deserialize)
        self.code_containers = CodeContainersOperations(self._client, self._config, self._serialize, self._deserialize)
        self.code_versions = CodeVersionsOperations(self._client, self._config, self._serialize, self._deserialize)
        self.component_containers = ComponentContainersOperations(self._client, self._config, self._serialize, self._deserialize)
        self.component_versions = ComponentVersionsOperations(self._client, self._config, self._serialize, self._deserialize)
        self.data_containers = DataContainersOperations(self._client, self._config, self._serialize, self._deserialize)
        self.data_versions = DataVersionsOperations(self._client, self._config, self._serialize, self._deserialize)
        self.datastores = DatastoresOperations(self._client, self._config, self._serialize, self._deserialize)
        self.environment_containers = EnvironmentContainersOperations(self._client, self._config, self._serialize, self._deserialize)
        self.environment_versions = EnvironmentVersionsOperations(self._client, self._config, self._serialize, self._deserialize)
        self.jobs = JobsOperations(self._client, self._config, self._serialize, self._deserialize)
        self.model_containers = ModelContainersOperations(self._client, self._config, self._serialize, self._deserialize)
        self.model_versions = ModelVersionsOperations(self._client, self._config, self._serialize, self._deserialize)
        self.online_endpoints = OnlineEndpointsOperations(self._client, self._config, self._serialize, self._deserialize)
        self.online_deployments = OnlineDeploymentsOperations(self._client, self._config, self._serialize, self._deserialize)
        self.operations = Operations(self._client, self._config, self._serialize, self._deserialize)
        self.workspaces = WorkspacesOperations(self._client, self._config, self._serialize, self._deserialize)
        self.usages = UsagesOperations(self._client, self._config, self._serialize, self._deserialize)
        self.virtual_machine_sizes = VirtualMachineSizesOperations(self._client, self._config, self._serialize, self._deserialize)
        self.quotas = QuotasOperations(self._client, self._config, self._serialize, self._deserialize)
        self.compute = ComputeOperations(self._client, self._config, self._serialize, self._deserialize)
        self.private_endpoint_connections = PrivateEndpointConnectionsOperations(self._client, self._config, self._serialize, self._deserialize)
        self.private_link_resources = PrivateLinkResourcesOperations(self._client, self._config, self._serialize, self._deserialize)
        self.workspace_connections = WorkspaceConnectionsOperations(self._client, self._config, self._serialize, self._deserialize)
        self.workspace_features = WorkspaceFeaturesOperations(self._client, self._config, self._serialize, self._deserialize)


    def _send_request(
        self,
        request,  # type: HttpRequest
        **kwargs  # type: Any
    ):
        # type: (...) -> HttpResponse
        """Runs the network request through the client's chained policies.

        >>> from azure.core.rest import HttpRequest
        >>> request = HttpRequest("GET", "https://www.example.org/")
        <HttpRequest [GET], url: 'https://www.example.org/'>
        >>> response = client._send_request(request)
        <HttpResponse: 200 OK>

        For more information on this code flow, see https://aka.ms/azsdk/python/protocol/quickstart

        :param request: The network request you want to make. Required.
        :type request: ~azure.core.rest.HttpRequest
        :keyword bool stream: Whether the response payload will be streamed. Defaults to False.
        :return: The response of your network call. Does not do error handling on your response.
        :rtype: ~azure.core.rest.HttpResponse
        """

        request_copy = deepcopy(request)
        request_copy.url = self._client.format_url(request_copy.url)
        return self._client.send_request(request_copy, **kwargs)

    def close(self):
        # type: () -> None
        self._client.close()

    def __enter__(self):
        # type: () -> AzureMachineLearningWorkspaces
        self._client.__enter__()
        return self

    def __exit__(self, *exc_details):
        # type: (Any) -> None
        self._client.__exit__(*exc_details)



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

from marshmallow import fields

from azure.ai.ml._schema.core.schema import PatchedSchemaMeta


class MLTableMetadataPathFileSchema(metaclass=PatchedSchemaMeta):
    file = fields.Str(
        metadata={"description": "This specifies path of file containing data."},
        required=True,
    )


class MLTableMetadataPathFolderSchema(metaclass=PatchedSchemaMeta):
    folder = fields.Str(
        metadata={"description": "This specifies path of folder containing data."},
        required=True,
    )


class MLTableMetadataPathPatternSchema(metaclass=PatchedSchemaMeta):
    pattern = fields.Str(
        metadata={
            "description": "This specifies a search pattern to allow globbing of files and folders containing data."
        },
        required=True,
    )



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------
import abc
import logging
import sys
from contextlib import contextmanager
from types import CodeType, FrameType, FunctionType, MethodType
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, Union

from azure.ai.ml._utils.utils import is_bytecode_optimization_enabled

logger = logging.getLogger(__name__)


class PersistentLocalsFunctionBuilder(abc.ABC):
    errors = {
        "not_callable": "func must be a function or a callable object",
        "conflict_argument": "Injected param name __self conflicts with function args {args}",
        "not_all_template_separators_used": "Not all template separators are used, "
        "please switch to a compatible version of Python.",
        "invalid_template": "Provided template functions are invalid in current environment, "
        "please switch to a compatible version (3.9 e.g.) of Python "
        "and/or check template functions.",
    }
    injected_param = "__self"

    @classmethod
    def make_error(cls, error_name: str, **kwargs) -> str:
        """Make error message with error_name and kwargs.

        :param error_name: A key from :attr:`~PersistentLocalsFunctionBuilder.errors`
        :type error_name: str
        :return: Formatted error message
        :rtype: str
        """
        return cls.errors[error_name].format(**kwargs)

    @abc.abstractmethod
    def _call(self, func, _all_kwargs) -> Tuple[Any, dict]:
        raise NotImplementedError()

    def call(self, func, _all_kwargs) -> Tuple[Any, dict]:
        """Get outputs and locals in calling func with _all_kwargs. Locals will be used to update node variable names.

        :param func: The function to execute.
        :type func: Union[FunctionType, MethodType]
        :param _all_kwargs: All kwargs to call self.func.
        :type _all_kwargs: typing.Dict[str, typing.Any]
        :return: A tuple of outputs and locals.
        :rtype: typing.Tuple[typing.Any, typing.Dict]
        """
        if isinstance(func, (FunctionType, MethodType)):
            pass
        elif hasattr(func, "__call__"):
            func = func.__call__
        else:
            raise TypeError(self.make_error("not_callable"))

        if self.injected_param in func.__code__.co_varnames:
            raise ValueError(self.make_error("conflict_argument", args=list(func.__code__.co_varnames)))

        return self._call(func, _all_kwargs)


class PersistentLocalsFunctionProfilerBuilder(PersistentLocalsFunctionBuilder):
    @staticmethod
    @contextmanager
    # pylint: disable-next=docstring-missing-return,docstring-missing-rtype
    def _replace_sys_profiler(profiler: Callable[[FrameType, str, Any], None]) -> Iterable[None]:
        """A context manager which replaces sys profiler to given profiler.

        :param profiler: The profile function.
            See https://docs.python.org/3/library/sys.html#sys.setprofile for more information
        :type profiler: Callable[[FrameType, str, Any], None]
        """
        original_profiler = sys.getprofile()
        sys.setprofile(profiler)
        try:
            yield
        finally:
            sys.setprofile(original_profiler)

    @staticmethod
    def _get_func_variable_tracer(
        _locals_data: Dict[str, Any], func_code: CodeType
    ) -> Callable[[FrameType, str, Any], None]:
        """Get a tracer to trace variable names in dsl.pipeline function.

        :param _locals_data: A dict to save locals data.
        :type _locals_data: dict
        :param func_code: An code object to compare if current frame is inside user function.
        :type func_code: CodeType
        :return: A tracing function
        :rtype: Callable[[FrameType, str, Any], None]
        """

        def tracer(frame: FrameType, event: str, arg: Any) -> None:  # pylint: disable=unused-argument
            if frame.f_code == func_code and event == "return":
                # Copy the locals of user's dsl function when it returns.
                _locals_data.update(frame.f_locals.copy())

        return tracer

    def _call(self, func, _all_kwargs):
        _locals = {}
        func_variable_profiler = self._get_func_variable_tracer(_locals, func.__code__)
        with self._replace_sys_profiler(func_variable_profiler):
            outputs = func(**_all_kwargs)
        return outputs, _locals


class PersistentLocalsFunction(object):
    def __init__(
        self,
        _func,
        *,
        _self: Optional[Any] = None,
        skip_locals: Optional[List[str]] = None,
    ):
        """
        :param _func: The function to be wrapped.
        :param _self: If original func is a method, _self should be provided, which is the instance of the method.
        :param skip_locals: A list of local variables to skip when saving the locals.
        """
        self._locals = {}
        self._self = _self
        # make function an instance method
        self._func = MethodType(_func, self)
        self._skip_locals = skip_locals

    def __call__(__self, *args, **kwargs):  # pylint: disable=no-self-argument
        # Use __self in case self is also passed as a named argument in kwargs
        __self._locals.clear()
        try:
            if __self._self:
                return __self._func(__self._self, *args, **kwargs)  # pylint: disable=not-callable
            return __self._func(*args, **kwargs)  # pylint: disable=not-callable
        finally:
            # always pop skip locals even if exception is raised in user code
            if __self._skip_locals is not None:
                for skip_local in __self._skip_locals:
                    __self._locals.pop(skip_local, None)

    @property
    def locals(self):
        return self._locals


def _source_template_func(mock_arg):
    return mock_arg


def _target_template_func(__self, mock_arg):
    try:
        return mock_arg
    finally:
        __self._locals = locals().copy()  # pylint: disable=protected-access


try:
    from bytecode import Bytecode, Instr, Label

    class PersistentLocalsFunctionBytecodeBuilder(PersistentLocalsFunctionBuilder):
        _template_separators = []
        _template_separators_before_body = []
        _template_separators_after_body = []
        _template_body = []
        _template_tail = None
        __initialized = False

        @classmethod
        def _split(cls, instructions, separator, n=-1):
            cur_start, index, result = 0, 0, []
            while index < len(instructions) - len(separator) + 1:
                if cls.is_instr_equal(instructions[index], separator[0]):
                    for i, template_body_instruction in enumerate(separator):
                        if not cls.is_instr_equal(instructions[index + i], template_body_instruction):
                            break
                    else:
                        result.append(instructions[cur_start:index])
                        cur_start = index + len(separator)
                        index += len(separator)
                        if len(result) == n:
                            break
                        continue
                index += 1
            result.append(instructions[cur_start:])
            if n != -1 and len(result) != n:
                msg = "can't split instructions into {} pieces with provided separators".format(n)
                raise ValueError(msg)
            return result

        @classmethod
        def _class_init_impl(cls):
            """Override this method to implement different template matching algorithm."""
            cls._template_separators_before_body, cls._template_separators_after_body = cls._split(
                cls.get_instructions(_source_template_func),
                separator=cls._get_mock_body_instructions(),
                n=2,
            )
            # use None to indicate the body
            cls._template_separators = (
                cls._template_separators_before_body + [None] + cls._template_separators_after_body
            )

            cls._template_body = cls._split_instructions_based_on_template(
                cls.get_instructions(_target_template_func),
                remove_mock_body=True,
            )
            cls._template_tail = cls._template_body.pop()
            if len(cls._template_body) != len(cls._template_body):
                raise ValueError(cls.make_error("invalid_template"))

        @classmethod
        def __class_init(cls):
            if cls.__initialized:
                return

            cls._class_init_impl()

            cls.__initialized = True

        def __init__(self):
            self.__class_init()

        # region methods depending on package bytecode
        @classmethod
        def get_instructions(cls, func):
            return list(Bytecode.from_code(func.__code__))

        @classmethod
        def is_instr_equal(cls, instr1: Instr, instr2: Instr) -> bool:
            if instr1 is None and instr2 is None:
                return True
            if instr1 is None or instr2 is None:
                return False
            if instr1.__class__ != instr2.__class__:
                return False
            if isinstance(instr1, Instr):
                if isinstance(instr1.arg, Label) and isinstance(instr2.arg, Label):
                    return True
                return instr1.opcode == instr2.opcode and instr1.arg == instr2.arg
            # objects like Label and TryBegin
            return True

        @classmethod
        def is_instructions_equal(cls, instructions1: List[Instr], instructions2: List[Instr]) -> bool:
            if len(instructions1) != len(instructions2):
                return False
            for instr1, instr2 in zip(instructions1, instructions2):
                if not cls.is_instr_equal(instr1, instr2):
                    return False
            return True

        def _create_code(self, instructions: List[Instr], base_func: Union[FunctionType, MethodType]) -> CodeType:
            """Create the base bytecode for the function to be generated.

            Will keep information of the function, such as name, globals, etc., but skip all instructions.

            :param instructions: The list of instructions. Used to replace the instructions in base_func
            :type instructions: List[Instr]
            :param base_func: A function that provides base metadata (name, globals, etc...). Instructions will not
                be kept
            :type base_func: Union[FunctionType, MethodType]
            :return: Generated code
            :rtype: CodeType
            """
            fn_code = Bytecode.from_code(base_func.__code__)
            fn_code.clear()
            fn_code.extend(instructions)
            fn_code.argcount += 1
            fn_code.argnames.insert(0, self.injected_param)
            return fn_code.to_code()

        @classmethod
        def _get_mock_body_instructions(cls):
            return [Instr("LOAD_FAST", "mock_arg")]

        # endregion

        @classmethod
        def _get_pieces(cls, instructions: List[Instr], separators: List[Instr]) -> List[List[Instr]]:
            """Split the instructions into pieces by the separators.
            Note that separators is a list of instructions. For example,
            instructions: [I3, I1, I2, I3, I1, I3, I1, I2, I3]
            separators: [I1, I2]
            result: [[I3], [I3, I1, I3], [I3]]

            :param instructions: The list of instructions to split
            :type instructions: List[instr]
            :param separators: The sequence of Instr to use as a delimiter
            :type separators: List[Instr]
            :return: A sublists of instructions that were delimited by separators
            :rtype: List[List[Instr]]
            """
            separator_iter = iter(separators)

            def get_next_separator():
                try:
                    while True:
                        separator = next(separator_iter)
                        if separator is not None:
                            return separator
                except StopIteration:
                    return None

            pieces = []
            last_piece = []
            cur_separator = get_next_separator()
            for instr in instructions:
                if cls.is_instr_equal(instr, cur_separator):
                    # skip the separator
                    pieces.append(last_piece)
                    cur_separator = get_next_separator()
                    last_piece = []
                else:
                    last_piece.append(instr)
            pieces.append(last_piece)

            if cur_separator is not None:
                raise ValueError(cls.make_error("not_all_template_separators_used"))

            return pieces

        @classmethod
        def _split_instructions_based_on_template(
            cls,
            instructions: List[Instr],
            *,
            remove_mock_body: bool = False,
        ) -> List[List[Instr]]:
            """Split instructions into several pieces by separators.
            For example, in Python 3.11, the template source instructions will be:

            .. code-block:: python

                [
                    Instr('RESUME', 0),  # initial instruction shared by all functions
                    Instr('LOAD_FAST', 'mock_arg'),  # the body execution instruction
                    Instr('RETURN_VALUE'),  # the return instruction shared by all functions
                ]

            Then the separators before body will be:

            .. code-block:: python

                [
                    Instr('RESUME', 0),
                ]

            And the separators after body will be:

            .. code-block:: python

                [
                    Instr('RETURN_VALUE'),
                ]

            For passed in instructions, we will split them with separators from beginning (the first RESUME) and
            with reversed_separators from end (the last RETURN_VALUE).

            :param instructions: The instructions to split
            :type instructions: List[instr]
            :keyword remove_mock_body: Whether to remove the mock body. Defaults to False
            :paramtype remove_mock_body: bool
            :return: The split instructions
            :rtype: List[List[Instr]]
            """
            if remove_mock_body:
                # this parameter should be set as True only when processing the template target function,
                # when we should ignore the mock body
                pieces = cls._get_pieces(
                    instructions, cls._template_separators_before_body + cls._get_mock_body_instructions()
                )
            else:
                pieces = cls._get_pieces(instructions, cls._template_separators_before_body)

            reversed_pieces = cls._get_pieces(reversed(pieces.pop()), reversed(cls._template_separators_after_body))

            while reversed_pieces:
                pieces.append(list(reversed(reversed_pieces.pop())))

            return pieces

        def _build_instructions(self, func: Union[FunctionType, MethodType]) -> List[Instr]:
            generated_instructions = []

            for template_piece, input_piece, separator in zip(
                self._template_body,
                self._split_instructions_based_on_template(self.get_instructions(func)),
                self._template_separators,
            ):
                generated_instructions.extend(template_piece)
                generated_instructions.extend(input_piece)
                if separator is not None:
                    generated_instructions.append(separator)
            generated_instructions.extend(self._template_tail)
            return generated_instructions

        def _build_func(self, func: Union[FunctionType, MethodType]) -> PersistentLocalsFunction:
            """Build a persistent locals function from the given function. Use bytecode injection to add try...finally
            statement around code to persistent the locals in the function.

            It will change the func bytecode in this way:

            .. code-block:: python

                def func(__self, *func_args):
                    try:
                       the func code...
                    finally:
                       __self.locals = locals().copy()

            You can get the locals in func by this code:

            .. code-block:: python

                builder = PersistentLocalsFunctionBuilder()
                persistent_locals_func = builder.build(your_func)
                # Execute your func
                result = persistent_locals_func(*args)
                # Get the locals in the func.
                func_locals = persistent_locals_func.locals

            :param func: The function to modify
            :type func: Union[FunctionType, MethodType]
            :return: The built persistent locals function
            :rtype: PersistentLocalsFunction
            """
            generated_func = FunctionType(
                self._create_code(self._build_instructions(func), func),
                func.__globals__,
                func.__name__,
                func.__defaults__,
                func.__closure__,
            )
            return PersistentLocalsFunction(
                generated_func,
                _self=func.__self__ if isinstance(func, MethodType) else None,
                skip_locals=[self.injected_param],
            )

        def _call(self, func, _all_kwargs) -> Tuple[Any, dict]:
            persistent_func = self._build_func(func)
            outputs = persistent_func(**_all_kwargs)
            return outputs, persistent_func.locals

except ImportError:
    # Fall back to the profiler implementation
    class PersistentLocalsFunctionBytecodeBuilder(PersistentLocalsFunctionProfilerBuilder):
        pass


def _get_persistent_locals_builder() -> PersistentLocalsFunctionBuilder:
    if is_bytecode_optimization_enabled():
        return PersistentLocalsFunctionBytecodeBuilder()
    return PersistentLocalsFunctionProfilerBuilder()


def get_outputs_and_locals(func, _all_kwargs):
    """Get outputs and locals from self.func. Locals will be used to update node variable names.

    :param func: The function to execute.
    :type func: Union[FunctionType, MethodType]
    :param _all_kwargs: All kwargs to call self.func.
    :type _all_kwargs: typing.Dict[str, typing.Any]
    :return: A tuple of outputs and locals.
    :rtype: typing.Tuple[typing.Dict, typing.Dict]
    """
    return _get_persistent_locals_builder().call(func, _all_kwargs)



from pathlib import Path

import pytest

from azure.ai.ml._file_utils.file_utils import traverse_up_path_and_find_file


@pytest.mark.unittest
@pytest.mark.core_sdk_test
class TestFileUtils:
    def test_traverse_up_path_and_find_file_multi_layers(self, tmp_path: Path):
        filename = "hello.txt"
        root = tmp_path
        filepath = root / filename
        filepath.write_text("content")

        start = root
        for i in range(10):
            start = start / f"sub{i}"
            start.mkdir()

        a = traverse_up_path_and_find_file(start, filename)
        assert a == str(filepath)

    def test_traverse_up_path_and_find_file(self, tmpdir):
        path = tmpdir.mkdir("sub")
        p = path.join("sub2")
        file_path = path.join("file.txt")
        file_path.write("content")
        ret = traverse_up_path_and_find_file(p, "file.txt")
        assert ret == file_path

    def test_traverse_up_path_and_find_file_with_dir_name(self, tmp_path: Path):
        filename = "hello.txt"
        root = tmp_path
        filepath = root / ".azure"
        filepath.mkdir()
        filepath = filepath / filename
        filepath.write_text("content")

        start = root
        for i in range(10):
            start = start / f"sub{i}"
            start.mkdir()

        a = traverse_up_path_and_find_file(start, filename, ".azure")
        assert a == str(filepath)



from pathlib import Path

from azure.ai.ml import Input, dsl, load_component
from azure.ai.ml.constants._common import AssetTypes, InputOutputModes
from azure.ai.ml.entities import PipelineJob

parent_dir = str(Path(__file__).parent)


def generate_dsl_pipeline() -> PipelineJob:
    # 1. Load component funcs
    get_data = load_component(source=parent_dir + "/get_data.yml")

    file_batch_inference = load_component(source=parent_dir + "/file_batch_inference.yml")
    file_batch_inference_duplicate = load_component(source=parent_dir + "/file_batch_inference.yml")
    tabular_batch_inference = load_component(source=parent_dir + "/tabular_batch_inference.yml")
    convert_data = load_component(source=parent_dir + "/convert_data.yml")

    # Construct pipeline
    @dsl.pipeline(default_compute="cpu-cluster")
    def parallel_in_pipeline(pipeline_job_data_path, pipeline_score_model):
        get_data_node = get_data(input_data=pipeline_job_data_path)
        get_data_node.outputs.file_output_data.type = AssetTypes.MLTABLE
        get_data_node.outputs.tabular_output_data.type = AssetTypes.MLTABLE

        file_batch_inference_node = file_batch_inference(job_data_path=get_data_node.outputs.file_output_data)
        file_batch_inference_node.inputs.job_data_path.mode = InputOutputModes.EVAL_MOUNT
        file_batch_inference_node.logging_level = "DEBUG"

        convert_data_node = convert_data(input_data=file_batch_inference_node.outputs.job_output_path)
        convert_data_node.outputs.file_output_data.type = AssetTypes.MLTABLE

        file_batch_inference_duplicate_node = file_batch_inference_duplicate(
            job_data_path=convert_data_node.outputs.file_output_data
        )
        file_batch_inference_duplicate_node.inputs.job_data_path.mode = InputOutputModes.EVAL_MOUNT
        file_batch_inference_duplicate_node.logging_level = "DEBUG"

        tabular_batch_inference_node = tabular_batch_inference(
            job_data_path=get_data_node.outputs.tabular_output_data, score_model=pipeline_score_model
        )
        tabular_batch_inference_node.inputs.job_data_path.mode = InputOutputModes.DIRECT
        tabular_batch_inference_node.logging_level = "DEBUG"

        return {
            "pipeline_job_out_file": file_batch_inference_duplicate_node.outputs.job_output_path,
            "pipeline_job_out_tabular": tabular_batch_inference_node.outputs.job_out_path,
        }

    pipeline = parallel_in_pipeline(
        pipeline_job_data_path=Input(
            type=AssetTypes.MLTABLE, path=parent_dir + "/dataset/", mode=InputOutputModes.RO_MOUNT
        ),
        pipeline_score_model=Input(
            path=parent_dir + "/model/", type=AssetTypes.URI_FOLDER, mode=InputOutputModes.DOWNLOAD
        ),
    )
    pipeline.outputs.pipeline_job_out_file.mode = "upload"
    pipeline.outputs.pipeline_job_out_tabular.type = "uri_file"

    return pipeline



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------
from typing import NoReturn

# pylint: disable=protected-access
from marshmallow import INCLUDE

from .._schema import NestedField
from ..entities._builders.control_flow_node import LoopNode
from ..entities._component.component_factory import component_factory
from ..entities._job.pipeline._load_component import pipeline_node_factory
from ._schema.command import CommandSchema, DistributedSchema, ParallelSchema
from ._schema.component import NodeType
from ._schema.node import HDInsightSchema, InternalBaseNodeSchema, ScopeSchema
from .entities import (
    Command,
    DataTransfer,
    Distributed,
    HDInsight,
    Hemera,
    InternalBaseNode,
    InternalComponent,
    Parallel,
    Pipeline,
    Scope,
    Starlite,
)
from .entities.spark import InternalSparkComponent

_registered = False


def _set_registered(value: bool):
    global _registered  # pylint: disable=global-statement
    _registered = value


def _enable_internal_components():
    create_schema_func = InternalComponent._create_schema_for_validation
    for _type in NodeType.all_values():
        component_factory.register_type(
            _type=_type,
            create_instance_func=lambda: InternalComponent.__new__(InternalComponent),
            create_schema_func=create_schema_func,
        )
    component_factory.register_type(
        _type=NodeType.SPARK,
        create_instance_func=lambda: InternalSparkComponent.__new__(InternalSparkComponent),
        create_schema_func=InternalSparkComponent._create_schema_for_validation,
    )


def _register_node(_type, node_cls, schema_cls):
    pipeline_node_factory.register_type(
        _type=_type,
        create_instance_func=lambda: node_cls.__new__(node_cls),
        load_from_rest_object_func=node_cls._from_rest_object,
        nested_schema=NestedField(schema_cls, unknown=INCLUDE),
    )


def enable_internal_components_in_pipeline(*, force=False) -> NoReturn:
    """Enable internal components in pipeline.

    :keyword force: Whether to force re-enable internal components. Defaults to False.
    :type force: bool
    :return: No return value.
    :rtype: None
    """
    if _registered and not force:
        return  # already registered

    _enable_internal_components()
    for _type in NodeType.all_values():
        # if we do not register node class for all node types, the only difference will be the type of created node
        # instance (Ae365exepool => InternalBaseNode). Not sure if this is acceptable.
        _register_node(_type, InternalBaseNode, InternalBaseNodeSchema)

    # redo the registration for those with specific runsettings
    _register_node(NodeType.DATA_TRANSFER, DataTransfer, InternalBaseNodeSchema)
    _register_node(NodeType.COMMAND, Command, CommandSchema)
    _register_node(NodeType.DISTRIBUTED, Distributed, DistributedSchema)
    _register_node(NodeType.PARALLEL, Parallel, ParallelSchema)
    _register_node(NodeType.HEMERA, Hemera, InternalBaseNodeSchema)
    _register_node(NodeType.STARLITE, Starlite, InternalBaseNodeSchema)
    _register_node(NodeType.SCOPE, Scope, ScopeSchema)
    _register_node(NodeType.HDI, HDInsight, HDInsightSchema)

    # register v2 style 1p only components
    _register_node(NodeType.HEMERA_V2, Hemera, InternalBaseNodeSchema)
    _register_node(NodeType.STARLITE_V2, Starlite, InternalBaseNodeSchema)
    _register_node(NodeType.SCOPE_V2, Scope, ScopeSchema)
    _register_node(NodeType.HDI_V2, HDInsight, HDInsightSchema)
    # Ae365exepool and AetherBridge have been registered to InternalBaseNode

    # allow using internal nodes in do-while loop
    LoopNode._extra_body_types = (Command, Pipeline)

    _set_registered(True)



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------
# pylint: disable=protected-access, too-many-locals

import os
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

from azure.ai.ml._restclient.v2023_04_01_preview.models import AmlToken, ManagedIdentity, UserIdentity
from azure.ai.ml.constants._common import AssetTypes
from azure.ai.ml.constants._component import ComponentSource
from azure.ai.ml.entities import Environment
from azure.ai.ml.entities._component.spark_component import SparkComponent
from azure.ai.ml.entities._inputs_outputs import Input, Output
from azure.ai.ml.entities._job.pipeline._component_translatable import ComponentTranslatableMixin
from azure.ai.ml.entities._job.spark_job_entry import SparkJobEntry
from azure.ai.ml.entities._job.spark_resource_configuration import SparkResourceConfiguration
from azure.ai.ml.exceptions import ErrorTarget, ValidationException

from .spark import Spark

SUPPORTED_INPUTS = [AssetTypes.URI_FILE, AssetTypes.URI_FOLDER, AssetTypes.MLTABLE]


def _parse_input(input_value: Union[Input, dict, str, bool, int, float]) -> Tuple:
    component_input = None
    job_input: Union[Input, dict, str, bool, int, float] = ""

    if isinstance(input_value, Input):
        component_input = Input(**input_value._to_dict())
        input_type = input_value.type
        if input_type in SUPPORTED_INPUTS:
            job_input = Input(**input_value._to_dict())
    elif isinstance(input_value, dict):
        # if user provided dict, we try to parse it to Input.
        # for job input, only parse for path type
        input_type = input_value.get("type", None)
        if input_type in SUPPORTED_INPUTS:
            job_input = Input(**input_value)
        component_input = Input(**input_value)
    elif isinstance(input_value, (str, bool, int, float)):
        # Input bindings are not supported
        component_input = ComponentTranslatableMixin._to_input_builder_function(input_value)
        job_input = input_value
    else:
        msg = f"Unsupported input type: {type(input_value)}, only Input, dict, str, bool, int and float are supported."
        raise ValidationException(message=msg, no_personal_data_message=msg, target=ErrorTarget.JOB)
    return component_input, job_input


def _parse_output(output_value: Union[Output, dict]) -> Tuple:
    component_output = None
    job_output: Union[Output, dict] = {}

    if isinstance(output_value, Output):
        component_output = Output(**output_value._to_dict())
        job_output = Output(**output_value._to_dict())
    elif not output_value:
        # output value can be None or empty dictionary
        # None output value will be packed into a JobOutput object with mode = ReadWriteMount & type = UriFolder
        component_output = ComponentTranslatableMixin._to_output(output_value)
        job_output = output_value
    elif isinstance(output_value, dict):  # When output value is a non-empty dictionary
        job_output = Output(**output_value)
        component_output = Output(**output_value)
    elif isinstance(output_value, str):  # When output is passed in from pipeline job yaml
        job_output = output_value
    else:
        msg = f"Unsupported output type: {type(output_value)}, only Output and dict are supported."
        raise ValidationException(message=msg, no_personal_data_message=msg, target=ErrorTarget.JOB)
    return component_output, job_output


def _parse_inputs_outputs(io_dict: Dict, parse_func: Callable) -> Tuple[Dict, Dict]:
    component_io_dict, job_io_dict = {}, {}
    if io_dict:
        for key, val in io_dict.items():
            component_io, job_io = parse_func(val)
            component_io_dict[key] = component_io
            job_io_dict[key] = job_io
    return component_io_dict, job_io_dict


def spark(
    *,
    experiment_name: Optional[str] = None,
    name: Optional[str] = None,
    display_name: Optional[str] = None,
    description: Optional[str] = None,
    tags: Optional[Dict] = None,
    code: Optional[Union[str, os.PathLike]] = None,
    entry: Union[Dict[str, str], SparkJobEntry, None] = None,
    py_files: Optional[List[str]] = None,
    jars: Optional[List[str]] = None,
    files: Optional[List[str]] = None,
    archives: Optional[List[str]] = None,
    identity: Optional[Union[Dict[str, str], ManagedIdentity, AmlToken, UserIdentity]] = None,
    driver_cores: Optional[int] = None,
    driver_memory: Optional[str] = None,
    executor_cores: Optional[int] = None,
    executor_memory: Optional[str] = None,
    executor_instances: Optional[int] = None,
    dynamic_allocation_enabled: Optional[bool] = None,
    dynamic_allocation_min_executors: Optional[int] = None,
    dynamic_allocation_max_executors: Optional[int] = None,
    conf: Optional[Dict[str, str]] = None,
    environment: Optional[Union[str, Environment]] = None,
    inputs: Optional[Dict] = None,
    outputs: Optional[Dict] = None,
    args: Optional[str] = None,
    compute: Optional[str] = None,
    resources: Optional[Union[Dict, SparkResourceConfiguration]] = None,
    **kwargs: Any,
) -> Spark:
    """Creates a Spark object which can be used inside a dsl.pipeline function or used as a standalone Spark job.

    :keyword experiment_name:  The name of the experiment the job will be created under.
    :paramtype experiment_name: Optional[str]
    :keyword name: The name of the job.
    :paramtype name: Optional[str]
    :keyword display_name: The job display name.
    :paramtype display_name: Optional[str]
    :keyword description: The description of the job. Defaults to None.
    :paramtype description: Optional[str]
    :keyword tags: The dictionary of tags for the job. Tags can be added, removed, and updated. Defaults to None.
    :paramtype tags: Optional[dict[str, str]]
    :keyword code: The source code to run the job. Can be a local path or "http:", "https:", or "azureml:" url
        pointing to a remote location.
    :type code: Optional[Union[str, os.PathLike]]
    :keyword entry: The file or class entry point.
    :paramtype entry: Optional[Union[dict[str, str], ~azure.ai.ml.entities.SparkJobEntry]]
    :keyword py_files: The list of .zip, .egg or .py files to place on the PYTHONPATH for Python apps.
        Defaults to None.
    :paramtype py_files: Optional[List[str]]
    :keyword jars: The list of .JAR files to include on the driver and executor classpaths. Defaults to None.
    :paramtype jars: Optional[List[str]]
    :keyword files: The list of files to be placed in the working directory of each executor. Defaults to None.
    :paramtype files: Optional[List[str]]
    :keyword archives: The list of archives to be extracted into the working directory of each executor.
        Defaults to None.
    :paramtype archives: Optional[List[str]]
    :keyword identity: The identity that the Spark job will use while running on compute.
    :paramtype identity: Optional[Union[
        dict[str, str],
        ~azure.ai.ml.entities.ManagedIdentityConfiguration,
        ~azure.ai.ml.entities.AmlTokenConfiguration,
        ~azure.ai.ml.entities.UserIdentityConfiguration]]
    :keyword driver_cores: The number of cores to use for the driver process, only in cluster mode.
    :paramtype driver_cores: Optional[int]
    :keyword driver_memory: The amount of memory to use for the driver process, formatted as strings with a size unit
        suffix ("k", "m", "g" or "t") (e.g. "512m", "2g").
    :paramtype driver_memory: Optional[str]
    :keyword executor_cores: The number of cores to use on each executor.
    :paramtype executor_cores: Optional[int]
    :keyword executor_memory: The amount of memory to use per executor process, formatted as strings with a size unit
        suffix ("k", "m", "g" or "t") (e.g. "512m", "2g").
    :paramtype executor_memory: Optional[str]
    :keyword executor_instances: The initial number of executors.
    :paramtype executor_instances: Optional[int]
    :keyword dynamic_allocation_enabled: Whether to use dynamic resource allocation, which scales the number of
        executors registered with this application up and down based on the workload.
    :paramtype dynamic_allocation_enabled: Optional[bool]
    :keyword dynamic_allocation_min_executors: The lower bound for the number of executors if dynamic allocation is
        enabled.
    :paramtype dynamic_allocation_min_executors: Optional[int]
    :keyword dynamic_allocation_max_executors: The upper bound for the number of executors if dynamic allocation is
        enabled.
    :paramtype dynamic_allocation_max_executors: Optional[int]
    :keyword conf: A dictionary with pre-defined Spark configurations key and values. Defaults to None.
    :paramtype conf: Optional[dict[str, str]]
    :keyword environment: The Azure ML environment to run the job in.
    :paramtype environment: Optional[Union[str, ~azure.ai.ml.entities.Environment]]
    :keyword inputs: A mapping of input names to input data used in the job. Defaults to None.
    :paramtype inputs: Optional[dict[str, ~azure.ai.ml.Input]]
    :keyword outputs: A mapping of output names to output data used in the job. Defaults to None.
    :paramtype outputs: Optional[dict[str, ~azure.ai.ml.Output]]
    :keyword args: The arguments for the job.
    :paramtype args: Optional[str]
    :keyword compute: The compute resource the job runs on.
    :paramtype compute: Optional[str]
    :keyword resources: The compute resource configuration for the job.
    :paramtype resources: Optional[Union[dict, ~azure.ai.ml.entities.SparkResourceConfiguration]]
    :return: A Spark object.
    :rtype: ~azure.ai.ml.entities.Spark

    .. admonition:: Example:

        .. literalinclude:: ../samples/ml_samples_spark_configurations.py
            :start-after: [START spark_function_configuration_1]
            :end-before: [END spark_function_configuration_1]
            :language: python
            :dedent: 8
            :caption: Configuring a SparkJob.

    .. admonition:: Example:

        .. literalinclude:: ../samples/ml_samples_spark_configurations.py
            :start-after: [START spark_function_configuration_2]
            :end-before: [END spark_function_configuration_2]
            :language: python
            :dedent: 8
            :caption: Configuring a SparkJob.

    .. admonition:: Example:

        .. literalinclude:: ../samples/ml_samples_spark_configurations.py
            :start-after: [START spark_dsl_pipeline]
            :end-before: [END spark_dsl_pipeline]
            :language: python
            :dedent: 8
            :caption: Building a Spark pipeline using the DSL pipeline decorator
    """

    inputs = inputs or {}
    outputs = outputs or {}
    component_inputs, job_inputs = _parse_inputs_outputs(inputs, parse_func=_parse_input)
    # job inputs can not be None
    job_inputs = {k: v for k, v in job_inputs.items() if v is not None}
    component_outputs, job_outputs = _parse_inputs_outputs(outputs, parse_func=_parse_output)
    component = kwargs.pop("component", None)

    if component is None:
        component = SparkComponent(
            name=name,
            display_name=display_name,
            tags=tags,
            description=description,
            code=code,
            entry=entry,
            py_files=py_files,
            jars=jars,
            files=files,
            archives=archives,
            driver_cores=driver_cores,
            driver_memory=driver_memory,
            executor_cores=executor_cores,
            executor_memory=executor_memory,
            executor_instances=executor_instances,
            dynamic_allocation_enabled=dynamic_allocation_enabled,
            dynamic_allocation_min_executors=dynamic_allocation_min_executors,
            dynamic_allocation_max_executors=dynamic_allocation_max_executors,
            conf=conf,
            environment=environment,
            inputs=component_inputs,
            outputs=component_outputs,
            args=args,
            _source=ComponentSource.BUILDER,
            **kwargs,
        )
    if isinstance(component, SparkComponent):
        spark_obj = Spark(
            experiment_name=experiment_name,
            name=name,
            display_name=display_name,
            tags=tags,
            description=description,
            component=component,
            identity=identity,
            driver_cores=driver_cores,
            driver_memory=driver_memory,
            executor_cores=executor_cores,
            executor_memory=executor_memory,
            executor_instances=executor_instances,
            dynamic_allocation_enabled=dynamic_allocation_enabled,
            dynamic_allocation_min_executors=dynamic_allocation_min_executors,
            dynamic_allocation_max_executors=dynamic_allocation_max_executors,
            conf=conf,
            inputs=job_inputs,
            outputs=job_outputs,
            compute=compute,
            resources=resources,
            **kwargs,
        )
    else:
        # when we load a remote job, component now is an arm_id, we need get entry from node level returned from
        # service
        spark_obj = Spark(
            experiment_name=experiment_name,
            name=name,
            display_name=display_name,
            tags=tags,
            description=description,
            component=component,
            identity=identity,
            driver_cores=driver_cores,
            driver_memory=driver_memory,
            executor_cores=executor_cores,
            executor_memory=executor_memory,
            executor_instances=executor_instances,
            dynamic_allocation_enabled=dynamic_allocation_enabled,
            dynamic_allocation_min_executors=dynamic_allocation_min_executors,
            dynamic_allocation_max_executors=dynamic_allocation_max_executors,
            conf=conf,
            inputs=job_inputs,
            outputs=job_outputs,
            compute=compute,
            resources=resources,
            entry=entry,
            py_files=py_files,
            jars=jars,
            files=files,
            archives=archives,
            args=args,
            **kwargs,
        )
    return spark_obj



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

import pytest
import yaml

from azure.ai.ml.entities._feature_set.feature_set_backfill_request import FeatureSetBackfillRequest
from azure.ai.ml.entities._load_functions import load_feature_set_backfill_request


@pytest.mark.unittest
@pytest.mark.data_experiences_test
class TestFeatureSetBackfillRequestSchema:
    def test_feature_set_backfill_request_load(self) -> None:
        test_path = "./tests/test_configs/feature_set/feature_set_backfill_request.yml"
        with open(test_path, "r") as f:
            target = yaml.safe_load(f)
        with open(test_path, "r") as f:
            backfill_request: FeatureSetBackfillRequest = load_feature_set_backfill_request(source=test_path)
        assert backfill_request.name == target["name"]
        assert backfill_request.version == target["version"]
        assert backfill_request.description == target["description"]
        assert backfill_request.data_status == target["data_status"]
        assert backfill_request.job_id == target["job_id"]
        assert backfill_request.tags is not None
        assert backfill_request.resource is not None
        assert backfill_request.spark_configuration is not None
        assert backfill_request.feature_window is not None
